//
//  MyQuat.c
//  TestApp4
//
//  Created by Yuji Nomura on 14/04/13.
//
//

#include "math.h"

/* ---------------- Quaternion operators ---------------- */

typedef struct { float w, x, y, z; } Quat4f;  /* of (w + xi + yj + zk) */
#define UNIT_QUAT4F_CONSTANT {1, 0, 0, 0}
const Quat4f UNIT_QUAT4F = UNIT_QUAT4F_CONSTANT;

void normalize_quat(Quat4f *q)
{
	float mag = sqrt(q->w * q->w + q->x * q->x + q->y * q->y + q->z * q->z);
	q->w /= mag;
	q->x /= mag;
	q->y /= mag;
	q->z /= mag;
}

void get_conjugate_quat(const Quat4f *q, Quat4f *result)
{
	Quat4f tmp;
	tmp.w =   q->w;
	tmp.x = - q->x;
	tmp.y = - q->y;
	tmp.z = - q->z;
	*result = tmp;
}

/* result.w <- p.w * q.w - p.v * q.v
   result.v <- p.w * q.v + q.w * p.v + p.v × q.v
   where v is vecter of (x,y,z) */
void mul_quat(const Quat4f *p, const Quat4f *q, Quat4f *result)
{
	Quat4f tmp;
	tmp.w = p->w * q->w - (p->x * q->x + p->y * q->y + p->z * q->z);
	tmp.x = p->w * q->x + q->w * p->x + (p->y * q->z - p->z * q->y);
	tmp.y = p->w * q->y + q->w * p->y + (p->z * q->x - p->x * q->z);
	tmp.z = p->w * q->z + q->w * p->z + (p->x * q->y - p->y * q->x);
	*result = tmp;
}

/* rotator quat: q = cos(th/2) + v * sin(th/2) */
void get_rot_quat(const float x, const float y, const float z, double angle, Quat4f *result)
{
	Quat4f tmp;
	double sin_angle = sin(angle/2);

	tmp.w = 0;
	tmp.x = x;
	tmp.y = y;
	tmp.z = z;
	normalize_quat(&tmp);

	result->w = cos(angle/2);
	result->x = tmp.x * sin_angle;
	result->y = tmp.y * sin_angle;
	result->z = tmp.z * sin_angle;
}

/* rotator from 2 quat:
   r.v <- p×q
   r.w <- cos(phi/2)
*/
void cross_product_quat(const Quat4f *p, const Quat4f *q, Quat4f *result)
{
	Quat4f tmp;
	tmp.w = 0;
	tmp.x = p->y * q->z - p->z * q->y;
	tmp.y = p->z * q->x - p->x * q->z;
	tmp.z = p->x * q->y - p->y * q->x;
	*result = tmp;
}

void get_rot_quat_from_quats(const Quat4f *p, const Quat4f *q, double angle, Quat4f *result)
{
	cross_product_quat(p, q, result);
	result->w = cos(angle/2);
}

void scale_quat_angle(Quat4f *q, float scale)
{
	double th = acos(q->w) * 2;
    if (th == 0) return;

	q->x /= sin(th/2);
	q->y /= sin(th/2);
	q->z /= sin(th/2);

	if (scale < 0){
		q->x = - q->x;
		q->y = - q->y;
		q->z = - q->z;
		scale = -scale;
	}

	th *= scale;
	q->x *= sin(th/2);
	q->y *= sin(th/2);
	q->z *= sin(th/2);
	q->w = cos(th/2);
}

void damp_quat_angle(Quat4f *q, float attenuation)
{
	double th = acos(q->w) * 2;

	q->x /= sin(th/2);
	q->y /= sin(th/2);
	q->z /= sin(th/2);
	
	th -= attenuation;
	if (th < 0.0){
		*q = UNIT_QUAT4F;
	}else{
		q->x *= sin(th/2);
		q->y *= sin(th/2);
		q->z *= sin(th/2);
		q->w = cos(th/2);
	}
}

/* convert quat to rotator matrix */
void quat_to_matrix(const Quat4f *q, float m[16])
{
	float x2 = q->x * q->x;
	float y2 = q->y * q->y;
	float z2 = q->z * q->z;
	float xy = q->x * q->y;
	float yz = q->y * q->z;
	float zx = q->z * q->x;
	float wx = q->w * q->x;
	float wy = q->w * q->y;
	float wz = q->w * q->z;
    
	m[0] = 1 - 2 * (y2 + z2);  m[4] = 2 * (xy + wz);	  m[8] = 2 * (zx - wy);	   m[12] = 0;
	m[1] = 2 * (xy - wz);	  m[5] = 1 - 2 * (z2 + x2);  m[9] = 2 * (yz + wx);	   m[13] = 0;
	m[2] = 2 * (zx + wy);	  m[6] = 2 * (yz - wx);	  m[10] = 1 - 2 * (x2 + y2);  m[14] = 0;
	m[3] = 0;				  m[7] = 0;				  m[11] = 0;				  m[15] = 1;
}

/* Slerp of quaternion:
                sin(a(1-t)) p + sin(a t) q
    result(t) = --------------------------
                          sin(a)
   where cos(a) = p * q
*/
void slerp_quat(const Quat4f *p, const Quat4f *q, float t, Quat4f *result)
{
	Quat4f tmp;
	Quat4f np, nq;  /* normalized quaternion */
	float a;
	double sin_a_1_t, sin_at, sin_a;

	if (t == 0.0){
		*result = *p;
	}
	else if (t == 1.0){
		*result = *q;
	}
	else {
		np = *p;
		nq = *q;
		np.w = 0;
		nq.w = 0;
		normalize_quat(&np);
		normalize_quat(&nq);
		a = acos(np.x * nq.x + np.y * nq.y + np.z * nq.z);
		sin_a_1_t = sin(a * (1-t));
		sin_at = sin(a * t);
		sin_a = sin(a);

		tmp.w = 0;
		tmp.x = (sin_a_1_t * np.x + sin_at * nq.x) / sin_a;
		tmp.y = (sin_a_1_t * np.y + sin_at * nq.y) / sin_a;
		tmp.z = (sin_a_1_t * np.z + sin_at * nq.z) / sin_a;
		*result = tmp;
	}
}

/* ---------------- End of quaternion operators ---------------- */

